/**
 * DESCRIPTION OF THE CLASS
 *
 * @author Martin Beseda
 * @author Martin Mrovec
 * @author Michal Kravčenko
 * @date 2017 - 2018
 */

#ifndef INC_4NEURO_NEURONLOGISTIC_H
#define INC_4NEURO_NEURONLOGISTIC_H

#include <cmath>
#include "Neuron.h"
#include "../constants.h"

class NeuronLogistic_d2:public Neuron, public IDifferentiable {
    friend class boost::serialization::access;
protected:
    template<class Archive>
    void serialize(Archive & ar, const unsigned int version){
        //TODO separate implementation to NeuronLogistic_d1.cpp!
        ar & boost::serialization::base_object<Neuron>(*this);
    };
public:

    Neuron* get_copy( );

    /**
     * Constructs the object of the Logistic neuron with activation function
     * f(x) = [(1 + a) e^(b-x) (1 + e^(b - x))^(-1) - 1]a e^(b - x) (1 + e^(b - x))^(-1 - a)
     * @param[in] a First coefficient, stored in activation_function_parameters[0]
     * @param[in] b Second coefficient, stored in activation_function_parameters[1]
     */
    explicit NeuronLogistic_d2(double a = 0.0, double b = 0.0);

    /**
     * Evaluates 'a*e^(b - x)*(1 + e^(b - x))^(-1 - a)' and stores the result into the 'state' property
     */
    void activate( ) override;

    /**
     * Calculates the partial derivative of the activation function
     * f(x) = [(1 + a) e^(b-x) (1 + e^(b - x))^(-1) - 1]a e^(b - x) (1 + e^(b - x))^(-1 - a)
     * @param[in] param_idx Index of the parameter to calculate derivative of
     * @return Partial derivative of the activation function according to the
     * 'param_idx'-th parameter.
     */
    double activation_function_eval_partial_derivative(int param_idx) override;

    /**
     * Calculates d/dx of  [(1 + a) e^(b-x) (1 + e^(b - x))^(-1) - 1]a e^(b - x) (1 + e^(b - x))^(-1 - a)
     * @return (a e^b (1 + e^(b - x))^(-a) (a^2 e^(2 b) + e^(2 x) - e^(b + x) - 3 a e^(b + x)))/(e^b + e^x)^3
     */
    double activation_function_eval_derivative( ) override;

};



class NeuronLogistic_d1:public Neuron, public IDifferentiable {
    friend class boost::serialization::access;
protected:
    template<class Archive>
    void serialize(Archive & ar, const unsigned int version){
        //TODO separate implementation to NeuronLogistic_d1.cpp!
        ar & boost::serialization::base_object<Neuron>(*this);
    };
public:

    Neuron* get_copy( );

    /**
     * Constructs the object of the Logistic neuron with activation function
     * f(x) = a*e^(b - x)*(1 + e^(b - x))^(-1 - a)
     * @param[in] a First coefficient, stored in activation_function_parameters[0]
     * @param[in] b Second coefficient, stored in activation_function_parameters[1]
     */
    explicit NeuronLogistic_d1(double a = 0.0, double b = 0.0);

    /**
     * Evaluates 'a*e^(b - x)*(1 + e^(b - x))^(-1 - a)' and stores the result into the 'state' property
     */
    void activate( ) override;

    /**
     * Calculates the partial derivative of the activation function
     * f(x) = a*e^(b - x)*(1 + e^(b - x))^(-1 - a)
     * @param[in] param_idx Index of the parameter to calculate derivative of
     * @return Partial derivative of the activation function according to the
     * 'param_idx'-th parameter.
     */
    double activation_function_eval_partial_derivative(int param_idx) override;

    /**
     * Calculates d/dx of  [a*e^(b - x)*(1 + e^(b - x))^(-1 - a)]
     * @return [[(1 + a) e^(b-x) (1 + e^(b - x))^(-1) - 1]a e^(b - x) (1 + e^(b - x))^(-1 - a)]
     */
    double activation_function_eval_derivative( ) override;

    /**
     * Returns a pointer to a Neuron with derivative as its activation function
     * @return
     */
    NeuronLogistic_d2* get_derivative() override;
};


class NeuronLogistic:public Neuron, public IDifferentiable {
    friend class boost::serialization::access;

protected:
    template<class Archive>
    void serialize(Archive & ar, const unsigned int version){
        //TODO separate implementation to NeuronLogistic.cpp!
        ar & boost::serialization::base_object<Neuron>(*this);
    };

public:

    Neuron* get_copy( );
    /**
     * Constructs the object of the Logistic neuron with activation function
     * f(x) = (1 + e^(-x + b))^(-a)
     * @param[in] a First coefficient, stored in activation_function_parameters[0]
     * @param[in] b Second coefficient, stored in activation_function_parameters[1]
     */
    explicit NeuronLogistic(double a = 0.0, double b = 0.0);

    /**
     * Evaluates '(1 + e^(-x + b))^(-a)' and stores the result into the 'state' property
     */
    void activate( ) override;

    /**
     * Calculates the partial derivative of the activation function
     * f(x) = (1 + e^(-x + b))^(-a)
     * @param[in] param_idx Index of the parameter to calculate derivative of
     * @return Partial derivative of the activation function according to the
     * 'param_idx'-th parameter.
     */
    double activation_function_eval_partial_derivative(int param_idx) override;
    /**
     * Calculates d/dx of (1 + e^(-x + b))^(-a)
     * @return a * e^(b - x) * [e^(b - x) + 1]^(-a)
     */
    double activation_function_eval_derivative( ) override;

    /**
     * Returns a pointer to a Neuron with derivative as its activation function
     * @return
     */
    NeuronLogistic_d1* get_derivative() override;
};


#endif //INC_4NEURO_NEURONLOGISTIC_H
