/**
 * DESCRIPTION OF THE FILE
 *
 * @author Michal Kravčenko
 * @date 22.7.18 -
 */

#include "DESolver.h"

MultiIndex::MultiIndex(unsigned int dimension) {
    this->dim = dimension;
    this->partial_derivatives_degrees.resize( this->dim );
    std::fill( this->partial_derivatives_degrees.begin(), this->partial_derivatives_degrees.end(), 0 );
}

void MultiIndex::set_partial_derivative(unsigned int index, unsigned int value) {
    this->partial_derivatives_degrees[ index ] = value;
}

std::vector<unsigned int>* MultiIndex::get_partial_derivatives_degrees() {
    return &this->partial_derivatives_degrees;
}

bool MultiIndex::operator<(const MultiIndex &rhs) const {
    if(dim < rhs.dim){ return true; }

    for(unsigned int i = 0; i < dim; ++i){
        if(partial_derivatives_degrees[i] < rhs.partial_derivatives_degrees[i]){
            return true;
        }
    }
    return false;
}

std::string MultiIndex::to_string( )const {
    std::string output;
    char buff[ 255 ];

    for( unsigned int i = 0; i < this->dim - 1; ++i){
        sprintf(buff, "%d, ", this->partial_derivatives_degrees[i]);
        output.append( buff );
    }
    sprintf(buff, "%d", this->partial_derivatives_degrees[this->dim - 1]);
    output.append( buff );

    return output;
}

unsigned int MultiIndex::get_degree() const{
    unsigned int output = 0;

    for( auto i: this->partial_derivatives_degrees ){
        output += i;
    }

    return output;
}


DESolver::DESolver( unsigned int n_equations, unsigned int n_inputs, unsigned int m, unsigned int n_outputs ) {

    if( m <= 0 || n_inputs <= 0 || n_outputs <= 0 || n_equations <= 0 ){
        throw std::invalid_argument("Parameters 'm', 'n_equations', 'n_inputs' and 'n_outputs' must be greater than zero!");
    }
    printf("Differential Equation Solver with %d equations\n--------------------------------------------------------------------------\n", n_equations);

    printf("Constructing NN structure representing the solution [%d input neurons][%d inner neurons][%d output neurons]...\n", n_inputs, m, n_outputs);

    this->dim_i = n_inputs;
    this->dim_o = n_outputs;
    this->dim_inn= m;
    this->n_equations = n_equations;

    this->solution = new NeuralNetwork();

    this->solution_inner_neurons = new std::vector<NeuronLogistic*>(0);
    this->solution_inner_neurons->reserve( m );

    /* input neurons */
    std::vector<size_t> input_set( this->dim_i );
    for( unsigned int i = 0; i < this->dim_i; ++i ){
        NeuronLinear *input_i = new NeuronLinear(0.0, 1.0);  //f(x) = x
        this->solution->add_neuron( input_i );
        input_set[i] = i;
    }
    this->solution->specify_input_neurons( input_set );

    /* output neurons */
    std::vector<size_t> output_set( this->dim_o );
    for( unsigned int i = 0; i < this->dim_o; ++i ){
        NeuronLinear *output_i = new NeuronLinear(0.0, 1.0);  //f(x) = x
        this->solution->add_neuron( output_i );
        output_set[i] = i + this->dim_i;
    }
    this->solution->specify_output_neurons( output_set );

    /* inner neurons */
    for(unsigned int i = 0; i < this->dim_inn; ++i){
        NeuronLogistic *inner_i = new NeuronLogistic(1.0, 0.0); //f(x) = 1.0 / (1.0 + e^(-x))
        this->solution_inner_neurons->push_back( inner_i );
        this->solution->add_neuron( inner_i );
    }

    /* connections between input neurons and inner neurons */
    size_t weight_idx;
    unsigned int dim_shift = this->dim_i + this->dim_o;
    for(unsigned int i = 0; i < this->dim_i; ++i){
        for(unsigned int j = 0; j < this->dim_inn; ++j){
            weight_idx = this->solution->add_connection_simple(i, dim_shift + j);
            printf("  adding a connection between input neuron %2d and inner neuron %2d, weight index %d\n", i, j, weight_idx);
        }
    }

    /* connections between inner neurons and output neurons */
    for(unsigned int i = 0; i < this->dim_inn; ++i){
        for(unsigned int j = 0; j < this->dim_o; ++j){
            weight_idx = this->solution->add_connection_simple(dim_shift + i, this->dim_i + j);
            printf("  adding a connection between inner neuron %2d and output neuron %2d, weight index %d\n", i, j, weight_idx);
        }
    }

    MultiIndex initial_mi(this->dim_i);

    this->map_multiindices2nn[initial_mi] = this->solution;

    this->differential_equations = new std::vector<NeuralNetworkSum*>(this->n_equations);

    for( unsigned int i = 0; i < this->n_equations; ++i ){
        this->differential_equations->at( i ) = new NeuralNetworkSum();
    }

    this->errors_functions_types = new std::vector<ErrorFunctionType >(this->n_equations);
    this->errors_functions_data_sets = new std::vector<DataSet*>(this->n_equations);

    printf("done\n\n");

}

DESolver::~DESolver() {

    if( this->solution_inner_neurons ){
        delete this->solution_inner_neurons;
        this->solution_inner_neurons = nullptr;
    }

    if( this->differential_equations ){
        delete this->differential_equations;
        this->differential_equations = nullptr;
    }

    if( this->errors_functions_types ){
        delete this->errors_functions_types;
        this->errors_functions_types = nullptr;
    }

    if( this->errors_functions_data_sets ){
        delete this->errors_functions_data_sets;
        this->errors_functions_data_sets = nullptr;
    }



}

void DESolver::add_to_differential_equation( unsigned int equation_idx, MultiIndex &alpha, double beta ) {

    if( equation_idx >= this->n_equations ){
        throw std::invalid_argument( "The provided equation index is too large!" );
    }

    unsigned int derivative_degree = alpha.get_degree( );

    if( derivative_degree > 2 ){
        throw std::invalid_argument("The supplied multi-index represents partial derivative of order higher than 2! (Valid degree is at most 2)\n");
    }

    /* retrieve indices of the variables according to which we perform the derivations ( applicable to any order, not just 2 or less )*/
    std::vector<unsigned int> partial_derivative_indices;
    partial_derivative_indices.reserve(derivative_degree);
    for( unsigned int i = 0; i < alpha.get_partial_derivatives_degrees()->size( ); ++i ){
        unsigned int degree = alpha.get_partial_derivatives_degrees()->at( i );

        while( degree > 0 ){

            partial_derivative_indices.push_back( i );
            degree--;

        }
    }



//    for( auto idx: map_multiindices2nn ){
////        printf("%s\n", idx.first.to_string().c_str());
//        printf("Is %s < %s? [%d]\n", idx.first.to_string().c_str(), alpha.to_string().c_str(), idx.first < alpha );
//        printf("Is %s < %s? [%d]\n", alpha.to_string().c_str(), idx.first.to_string().c_str(), alpha < idx.first );
//    }

    NeuralNetwork *new_net = nullptr;
    /* we check whether the new multi-index is already present */
    if(map_multiindices2nn.find( alpha ) != map_multiindices2nn.end()){
        new_net = map_multiindices2nn[ alpha ];
        this->differential_equations->at( equation_idx )->add_network( new_net, beta );
        printf("Adding an existing partial derivative (multi-index: %s) to equation %d with coefficient %f\n", alpha.to_string().c_str(), equation_idx, beta);
        return;
    }
    printf("Adding a new partial derivative (multi-index: %s) to equation %d with coefficient %f\n", alpha.to_string().c_str(), equation_idx, beta);

    /* we need to construct a new neural network */
     new_net = new NeuralNetwork( );
     new_net->set_weight_array( this->solution->get_weights_ptr() );

    /* input neurons */
    std::vector<size_t> input_set( this->dim_i );
    for( unsigned int i = 0; i < this->dim_i; ++i ){
        NeuronLinear *input_i = new NeuronLinear(0.0, 1.0);  //f(x) = x
        new_net->add_neuron( input_i );
        input_set[i] = i;
    }
    new_net->specify_input_neurons( input_set );

    /* output neurons */
    std::vector<size_t> output_set( this->dim_o );
    for( unsigned int i = 0; i < this->dim_o; ++i ){
        NeuronLinear *output_i = new NeuronLinear(0.0, 1.0);  //f(x) = x
        new_net->add_neuron( output_i );
        output_set[i] = i + this->dim_i;
    }
    new_net->specify_output_neurons( output_set );

    /* the new partial derivative has degree of at least one */
    for( unsigned int i = 0; i < this->dim_inn; ++i ){
        if( derivative_degree == 1 ){
            NeuronLogistic_d1 *new_inn_neuron = this->solution_inner_neurons->at( i )->get_derivative( );
            new_net->add_neuron( new_inn_neuron );
        }
        else if( derivative_degree == 2 ){
            NeuronLogistic_d2 *new_inn_neuron = this->solution_inner_neurons->at( i )->get_derivative( )->get_derivative( );
            new_net->add_neuron( new_inn_neuron );
        }
    }

    /* identity neurons serving as a 'glue'*/
    for(unsigned int i = 0; i < derivative_degree * this->dim_o * this->dim_inn; ++i){
        NeuronLinear *new_neuron = new NeuronLinear(0.0, 1.0);  //f(x) = x
        new_net->add_neuron( new_neuron );
    }

    /* connections between input neurons and inner neurons */
    unsigned int dim_shift = this->dim_i + this->dim_o;
    size_t weight_idx = 0;
    for(unsigned int i = 0; i < this->dim_i; ++i){
        for(unsigned int j = 0; j < this->dim_inn; ++j){
            printf("  adding a connection between input neuron %2d and inner neuron %2d, weight index %d\n", i, j, weight_idx);
            new_net->add_connection_simple(i, dim_shift + j, weight_idx);
            weight_idx++;
        }
    }

    /* connections between inner neurons and the first set of 'glueing' neurons */
//    unsigned int glue_shift = dim_shift + this->dim_inn;
    unsigned int glue_shift = dim_shift + this->dim_inn;
//    size_t weight_idx_mem = weight_idx;
    for(unsigned int i = 0; i < this->dim_inn; ++i){
        for(unsigned int j = 0; j < this->dim_o; ++j){
            printf("  adding a connection between inner neuron %2d and glue neuron %2d, weight index %d\n", i, i * this->dim_o + j, weight_idx);
            new_net->add_connection_simple(dim_shift + i, glue_shift + i * this->dim_o + j, weight_idx );
            weight_idx++;
        }
    }

    if( derivative_degree == 1 ){
        /* connections from the first set of glueing neurons to the output neurons */
        unsigned int pi = partial_derivative_indices[0];

        for(unsigned int i = 0; i < this->dim_inn; ++i){
            unsigned int local_weight_idx = pi * this->dim_inn + i;

            for(unsigned int j = 0; j < this->dim_o; ++j){
                printf("  adding a connection between glue neuron %2d and output neuron %2d, weight index %d\n", i * this->dim_o + j, j, local_weight_idx);
                new_net->add_connection_simple(glue_shift + i * this->dim_o + j, j + this->dim_i, local_weight_idx );
            }
        }
    }
    else if( derivative_degree == 2 ){
        /* connections from the first set of glueing neurons towards the second set of glueing neurons */
        unsigned int pi = partial_derivative_indices[0];
        unsigned int glue_shift_2 = this->dim_inn + this->dim_i + this->dim_o + this->dim_inn * this->dim_o;

        for(unsigned int i = 0; i < this->dim_inn; ++i){
            unsigned int local_weight_idx = pi * this->dim_inn + i;

            for(unsigned int j = 0; j < this->dim_o; ++j){
                printf("  adding a connection between glue neuron[1] %2d and glue neuron[2] %2d, weight index %d\n", i * this->dim_o + j, i * this->dim_o + j, local_weight_idx);
                new_net->add_connection_simple(glue_shift + i * this->dim_o + j, glue_shift_2 + i * this->dim_o + j, local_weight_idx );
            }
        }

        /* connections from the second set of glueing neurons towards the output neurons */
        pi = partial_derivative_indices[1];

        for(unsigned int i = 0; i < this->dim_inn; ++i){
            unsigned int local_weight_idx = pi * this->dim_inn + i;

            for(unsigned int j = 0; j < this->dim_o; ++j){
                printf("  adding a connection between glue neuron[2] %2d and output neuron %2d, weight index %d\n", i * this->dim_o + j, j, local_weight_idx);
                new_net->add_connection_simple(glue_shift_2 + i * this->dim_o + j, this->dim_i + j, local_weight_idx );
            }
        }
    }
    map_multiindices2nn[ alpha ] = new_net;

    this->differential_equations->at( equation_idx )->add_network( new_net, beta );
}

void DESolver::set_error_function(unsigned int equation_idx, ErrorFunctionType F, DataSet *conditions) {
    if( equation_idx >= this->n_equations ){
        throw std::invalid_argument("The parameter 'equation_idx' is too large! It exceeds the number of differential equations.");
    }

    this->errors_functions_types->at( equation_idx ) = F;
    this->errors_functions_data_sets->at( equation_idx ) = conditions;
}

void DESolver::solve_via_particle_swarm(double *domain_bounds, double c1, double c2, double w,
                                          unsigned int n_particles, unsigned int max_iters, double gamma,
                                          double epsilon, double delta) {

    NeuralNetwork *nn;
    DataSet *ds;

    /* DEFINITION OF THE PARTIAL ERROR FUNCTIONS */
    std::vector<ErrorFunction*> error_functions( this->n_equations );
    for(unsigned int i = 0; i < this->n_equations; ++i ){
        nn = this->differential_equations->at( i );
        ds = this->errors_functions_data_sets->at( i );

        if( this->errors_functions_types->at( i ) == ErrorFunctionType::ErrorFuncMSE ){
            error_functions[i] = new MSE( nn, ds );
        }
        else{
            //default
            error_functions[i] = new MSE( nn, ds );
        }
    }

    /* DEFINITION OF THE GLOBAL ERROR FUNCTION */
    ErrorSum total_error;
    for(unsigned int i = 0; i < this->n_equations; ++i ) {
        total_error.add_error_function( error_functions[i], 1.0 );
    }

    ParticleSwarm swarm_01(&total_error, domain_bounds, c1, c2, w, n_particles, max_iters);

    this->solution->randomize_weights();

    swarm_01.optimize(gamma, epsilon, delta);

}

NeuralNetwork* DESolver::get_solution() {
    return this->solution;
}

void DESolver::eval_equation( unsigned int equation_idx, double *weight_values, std::vector<double> &input ) {
    std::vector<double> output(this->dim_o);

    this->differential_equations->at( equation_idx )->eval_single( input, output, weight_values );

    printf("Input: ");
    for( auto e: input ){
        printf("%f, ", e);
    }
    printf("\nOutput: ");
    for( auto e: output ){
        printf("%f, ", e);
    }
    printf("\n");
}