/**
 * File containing methods for quick & simple formulation of PDEs as a system of Neural Networks with shared weights
 *
 * @author Michal Kravčenko
 * @date 22.7.18 -
 */

 //TODO incorporate uncertainities as coefficients in NeuralNetworkSum or ErrorSum

#ifndef INC_4NEURO_PDESOLVER_H
#define INC_4NEURO_PDESOLVER_H

#include <map>
#include "../DataSet/DataSet.h"
#include "../Network/NeuralNetwork.h"
#include "../Network/NeuralNetworkSum.h"
#include "../ErrorFunction/ErrorFunctions.h"
#include "../Neuron/NeuronLinear.h"
#include "../Neuron/NeuronLogistic.h"
#include "../LearningMethods/ParticleSwarm.h"

/**
 * class representing a multi-index of partial derivatives
 */
class MultiIndex{
private:
    /**
     * total number of variables
     */
    unsigned int dim;

    /**
     * a vector containing degrees of partial derivatives with respect to each variable
     */
    std::vector<unsigned int> partial_derivatives_degrees;

public:
    /**
     *
     * @param dimension
     */
    MultiIndex(unsigned int dimension);


    /**
     *
     * @param index
     * @param value
     */
    void set_partial_derivative(unsigned int index, unsigned int value);

    /**
     *
     * @return
     */
    std::vector<unsigned int>* get_partial_derivatives_degrees( ) ;


    /**
     *
     * @param rhs
     * @return
     */
    bool operator <(const MultiIndex& rhs) const;

    /**
     *
     * @return
     */
    std::string to_string( ) const ;

    /**
     *
     * @return
     */
    unsigned int get_degree( ) const ;
};



class DESolver {
private:

    /* Mapping between used multiindices of partial derivatices and already defined NNs (in order to not define
     * the same NN multiple times )*/
    std::map<MultiIndex, NeuralNetwork*> map_multiindices2nn;

    /* A list of the differential equations */
    std::vector<NeuralNetworkSum*> * differential_equations = nullptr;

    /* Error functions for differential equations */
    std::vector<ErrorFunctionType> * errors_functions_types = nullptr;
    std::vector<DataSet*> * errors_functions_data_sets = nullptr;

    /* NN as the unknown function */
    NeuralNetwork * solution = nullptr;


    /* auxilliary variables */
    std::vector<NeuronLogistic*> *solution_inner_neurons = nullptr;
    unsigned int dim_i = 0, dim_o = 0, dim_inn = 0, n_equations = 0;

public:
    /**
     * The attempted solution will contain 'm' inner neurons between the input neurons and the output neuron
     * @param n_equations
     * @param n_inputs
     * @param m
     * @param n_outputs
     */
    DESolver( unsigned int n_equations, unsigned int n_inputs, unsigned int m, unsigned int n_outputs );

    /**
     * default destructor
     */
    ~DESolver( );

    /**
     * Adds a new summand multiplied by 'beta' into the 'equation_idx'-th differential equation
     * @param equation_idx
     * @param alpha
     * @param beta
     */
    void add_to_differential_equation( unsigned int equation_idx, MultiIndex &alpha, double beta );

    /**
     * Sets the error function for the differential equation with the corresponding index
     * @param equation_idx
     * @param F
     * @param conditions
     */
    void set_error_function(unsigned int equation_idx, ErrorFunctionType F, DataSet *conditions);


     /**
      * solves the PDE with its boundary conditions via the particle swarm algorithm
      * @param domain_bounds
      * @param c1
      * @param c2
      * @param w
      * @param n_particles
      * @param max_iters
      * @param gamma
      * @param epsilon
      * @param delta
      * @return
      */
    void solve_via_particle_swarm(
            double * domain_bounds,
            double c1,
            double c2,
            double w,
            unsigned int n_particles,
            unsigned int max_iters,
            double gamma,
            double epsilon,
            double delta
            );

    /**
     * returns the pointer to the object representing the solution y(x_1, x_2, ..., x_n)
     * @return
     */
    NeuralNetwork* get_solution( );

    /**
     * For testing purposes only
     */
     void eval_equation( unsigned int equation_idx, double * weight_values, std::vector<double> &input );
};


#endif //INC_4NEURO_PDESOLVER_H
